# Authors: Michael Sander, Pierre Ablin
# License: MIT

__version__ = '0.8.dev'

from .momentum_net import MomentumNet, MomentumMemory, Mom  # noqa
from .resnet_to_momentumnet import transform # noqa
